#include "StdAfx.h"
#include "SampleUtils.h"


SampleUtils::SampleUtils(void)
{
}

SampleUtils::~SampleUtils(void)
{
}

/*
* Name: getAssocNetwork
*
* Parameters:
* [INPUT]bool createIfNotExist: boolean indicating if assoc network should be created if it does not exist
* [INPUT]AcDbDatabase* pDatabase: The database from which the network is sought
*
* Return: Returns the objectid of the associative network
*/

AcDbObjectId SampleUtils::getAssocNetwork(bool createIfNotExist,
										   AcDbDatabase* pDatabase)
{
	Acad::ErrorStatus eStatus = eOk;
	if (pDatabase == NULL)
	{
		pDatabase = acdbHostApplicationServices()->workingDatabase();
	    if (!pDatabase)
		    return AcDbObjectId::kNull;
	}

	//get the existing network
	return AcDbAssocNetwork::getInstanceFromObject(pDatabase->currentSpaceId(), createIfNotExist);
}

/*
* Name: getConstraintGroup
*
* Parameters:
* [INPUT]AcDbObjectId btrId: The objectid of block table record in which constraints are being applied
* [INPUT]bool  createIfDoesNotExist: boolen indicating if the constraint group should be created if it does not exist already
* [INPUT]const AcString& dictionaryKey: Name of the dictionary entry in the NOD that represents the top network object under which
*                                all constraint object are built as a tree structure
*
* Return: Returns the objectid of the constraint group
*/

AcDbObjectId SampleUtils::getConstraintGroup(AcDbObjectId btrId,
                                              bool  createIfDoesNotExist,
                                              const AcString& dictionaryKey)
{
	// Find the top network
	AcDbDatabase* pDatabase = btrId.database();
	AcDbObjectId idConstrGroup = AcDbObjectId::kNull;
    AcDbObjectId idTopNetwork = 
        AcDbAssocNetwork::getInstanceFromDatabase(pDatabase,
                                                  createIfDoesNotExist,
                                                  dictionaryKey);

    if (idTopNetwork.isNull())
	  {
        ASSERT(!createIfDoesNotExist);
        return idConstrGroup; // AcDbObjectId::kNull;
    }

	//  Get the network associated with the block table record
    AcDbObjectId idNetwork = 
      AcDbAssocNetwork::getInstanceFromObject(btrId,
      createIfDoesNotExist,
      true);

    if (idNetwork.isNull())
    {
        ASSERT(!createIfDoesNotExist);
        return idConstrGroup; // AcDbObjectId::kNull;
    }

	//
    AcDbSmartObjectPointer<AcDbAssocNetwork> pTopNetwork(idTopNetwork, AcDb::kForRead);
    if (!eOkVerify(pTopNetwork.openStatus()))
    {
        return idConstrGroup; // AcDbObjectId::kNull;
    }

    if (!pTopNetwork->getActions().contains(idNetwork))
    {
        pTopNetwork.open(idTopNetwork, AcDb::kForWrite);
        if (!eOkVerify(pTopNetwork.openStatus()))
          return idConstrGroup; // AcDbObjectId::kNull;

        pTopNetwork->addAction(idNetwork, false);
    }

     // get the current UCS plane
     AcGeMatrix3d ucsMatrix;
     acedGetCurrentUCS(ucsMatrix);
     AcGePoint3d origin;
     AcGeVector3d xAxis, yAxis, zAxis;
     ucsMatrix.getCoordSystem(origin, xAxis, yAxis, zAxis);
     AcGePlane currentPlane(origin, xAxis, yAxis);

    // Try to find the constraint group in the associative network
	 if (idConstrGroup.isNull() && createIfDoesNotExist)
	 {
        AcDbObjectPointer<AcDbAssocNetwork> pNetwork(idNetwork, kForRead);
        if (eOk != pNetwork.openStatus()) {
            ASSERT(false);
            return idConstrGroup; // AcDbObjectId::kNull;
        }

		AcRxClass *pBaseClass = AcDbAssoc2dConstraintGroup::desc();
        const AcDbObjectIdArray& actionsInNetwork = pNetwork->getActions();
        for (int nCount = 0; nCount < actionsInNetwork.length(); nCount++)
        {
			const AcDbObjectId& idAction = actionsInNetwork[nCount];
            if (idAction == AcDbObjectId::kNull)
                continue;
			if (actionsInNetwork[nCount].objectClass() == NULL || 
				!actionsInNetwork[nCount].objectClass()->isDerivedFrom(pBaseClass))
				continue;
			AcDbObjectPointer<AcDbAssocAction> pAction(idAction, kForRead);
		    if (eOk != pAction.openStatus())
				continue;
			const AcDbAssoc2dConstraintGroup* const pConstGrp = 
				static_cast<AcDbAssoc2dConstraintGroup*>(pAction.object());
			ASSERT(pConstGrp);
			if (!pConstGrp)
				continue;

			AcGePlane constGrpPlane = pConstGrp->getWorkPlane();
			if (constGrpPlane.isCoplanarTo(currentPlane))
			{
                // We have an constarint group which is built on same UCS
				return idAction;
			}
		}
    }

	 // Create 2D constraint group if it does not exist
    if (idConstrGroup.isNull() && createIfDoesNotExist)
    {
        AcDbObjectPointer<AcDbAssocNetwork> pNetwork(idNetwork, kForWrite);
        if (eOk != pNetwork.openStatus()) {
            ASSERT(false);
            return AcDbObjectId::kNull;
        }

        AcDbAssoc2dConstraintGroup* const pConstrGroup =
            new AcDbAssoc2dConstraintGroup(currentPlane);

        AcDbDatabase *pDatabase = pNetwork->database();
        if (eOk != pDatabase->addAcDbObject(idConstrGroup, pConstrGroup))
		{
            assert(false);
            delete pConstrGroup;
            return AcDbObjectId::kNull;
        }

        pConstrGroup->close();

        if (eOk != pNetwork->addAction(idConstrGroup, true)) {
            ASSERT(false);
            return AcDbObjectId::kNull;
        }
    }

    return idConstrGroup;

}

/*
* Name: getConstraintGroup
*
* Parameters:
* [INPUT]AcDbDatabase  *pDatabase: The database from which the constraint group is sought
* [INPUT]bool  createIfDoesNotExiste: boolean indicating is the constraint group should be created if it does not exist
*
* Return: ObjectId of the constraint group
*/
AcDbObjectId 
SampleUtils::getConstraintGroup(AcDbDatabase  *pDatabase,
											  bool  createIfDoesNotExist)
{
	Acad::ErrorStatus err = eOk;

	AcGeMatrix3d ucsMatrix;

	AcGePoint3d origin(0,0,0);
	AcGeVector3d xAxis(1,0,0), yAxis(0,1,0), zAxis(0,0,1);
	ucsMatrix.getCoordSystem(origin, xAxis, yAxis, zAxis);
	AcGePlane currentPlane(origin, xAxis, yAxis);

	if (pDatabase == NULL)
	{
		pDatabase = acdbHostApplicationServices()->workingDatabase();
		if (!pDatabase)
			return AcDbObjectId::kNull;
	}

	//get existing network
	AcDbObjectId networkId = getAssocNetwork(createIfDoesNotExist, pDatabase);

	AcDbObjectId consGrpId = AcDbObjectId::kNull;

	//try to find a constraint group in current network
	{
		AcDbObjectPointer<AcDbAssocNetwork> pNetwork(networkId, kForRead);
		if (pNetwork.openStatus() != eOk)
			return consGrpId;

		AcRxClass* pBaseClass = AcDbAssoc2dConstraintGroup::desc();
		const AcDbObjectIdArray& actionsInNetwork = pNetwork->getActions();
		for (int nCount = 0; nCount < actionsInNetwork.length(); nCount++)
		{
			const AcDbObjectId& actionId = actionsInNetwork[nCount];
			if (actionId == AcDbObjectId::kNull)
				continue;

			if (actionId.objectClass() == NULL ||
				!actionId.objectClass()->isDerivedFrom(pBaseClass))
				continue;

			AcDbObjectPointer<AcDbAssocAction> pAction(actionId, kForRead);
			if (pAction.openStatus() != eOk)
				continue;

			const AcDbAssoc2dConstraintGroup* const pConsGrp = 
				static_cast<AcDbAssoc2dConstraintGroup*>(pAction.object());
			if (!pConsGrp)
				continue;

			AcGePlane consGrpPlane = pConsGrp->getWorkPlane();
			if (consGrpPlane.isCoplanarTo(currentPlane))
				return actionId;
		}
	}

	if (consGrpId.isNull() && createIfDoesNotExist)
	{
		AcDbObjectPointer<AcDbAssocNetwork> pNetwork(networkId, kForWrite);
		if (pNetwork.openStatus() != eOk)
			return AcDbObjectId::kNull;

		AcDbAssoc2dConstraintGroup* const pConsGrp = new AcDbAssoc2dConstraintGroup(currentPlane);

		pDatabase = pNetwork->database();
		if (eOk != pDatabase->addAcDbObject(consGrpId, pConsGrp))
		{
			assert(false);
			delete pConsGrp;
			return AcDbObjectId::kNull;
		}

		pConsGrp->close();

		if (eOk != pNetwork->addAction(consGrpId, true))
		{
			assert(false);
			return AcDbObjectId::kNull;
		}
	}

	return consGrpId;
}

/*
* Name: getSubentIDs
*
* Parameters:
* [INPUT]const AcDbEntity* pEntity: The entity whose subentities are being queried
* [INPUT]AcDb::SubentType subentType: The type of subentity being queried for
* [OUTPUT]AcArray<AcDbSubentId>& allSubentIds: The output array containing the list of subentites being queried for
*
* Return: Error status
*/
Acad::ErrorStatus SampleUtils::getSubentIDs(const AcDbEntity*      pEntity,
                        AcDb::SubentType       subentType,
                        AcArray<AcDbSubentId>& allSubentIds)
{
	Acad::ErrorStatus err;

	AcDbAssocPersSubentIdPE* const pAssocPersSubentIdPE1
		= AcDbAssocPersSubentIdPE::cast(pEntity->queryX(AcDbAssocPersSubentIdPE::desc()));
		
		err = pAssocPersSubentIdPE1->getAllSubentities(pEntity,
												   subentType,
												   allSubentIds);

		return err;
}


/*
* Name: addConstrainedGeometry
*
* Parameters:
* [INPUT]AcDbAssoc2dConstraintGroup *pConstGrp: The 2D constraint group to which the subentities are added
* [INPUT]const AcDbFullSubentPathArray& aPaths: The array of subentities to add
*
* Return: Error status
*/
Acad::ErrorStatus SampleUtils::addConstrainedGeometry(AcDbAssoc2dConstraintGroup *pConstGrp, const AcDbFullSubentPathArray& aPaths)
{
	Acad::ErrorStatus err=Acad::eOk;

	AcConstrainedGeometry* pConsGeom = NULL;

	int i = 0;
    //Add the relevent subentities as the constrained geometry
	while(i < aPaths.length())
	{
		err = pConstGrp->addConstrainedGeometry(aPaths[i++], pConsGeom);
	}
	
	return err;
}

void SampleUtils::evaluateNetwork()
{
	//Evaluate the network
	CCustomSampleCallback callback;

	AcDbDatabase *pDatabase = acdbHostApplicationServices()->workingDatabase();

	AcDbObjectId idNetwork = 
	AcDbAssocNetwork::getInstanceFromDatabase(pDatabase,
											  true,
											  DICTIONARYKEY);

	AcDbObjectPointer<AcDbAssocNetwork> pNetwork(idNetwork, kForWrite);
	ASSERT (eOk == pNetwork.openStatus());

	pNetwork->evaluate(&callback);

}

/*
* Name: getNameAndExpressionFromDimension
*
* Parameters:
* [INPUT]const AcDbObjectId &dimId: The objectID of the dimension for which the name and expression are sought
* [INPUT]const AcDbAssocConstraintType &constraintType: The constraint type for which the name and expression are sought
* [OUTPUT]AcString &name: The resulting name
* [OUTPUT]AcString &expression: The resulting expression
*
* Return: Error status
*/
Acad::ErrorStatus
SampleUtils::getNameAndExpressionFromDimension(const AcDbObjectId &dimId,
															 const AcDbAssocConstraintType &constraintType,
															 AcString &name,
															 AcString &expression)
{
	Acad::ErrorStatus err = eOk;
	AcString namePrefix, dfltName, dfltExpression;
	name.setEmpty();
	expression.setEmpty();

	// Choose the name prefix based on the type of constraint
	switch (constraintType)
	{
	case kDistanceAssocConstraintType:
	case kHorizontalDistanceAssocConstraintType:
	case kVerticalDistanceAssocConstraintType:
		namePrefix = ACRX_T("d");
		break;

	case kAngle0AssocConstraintType:
	case kAngle1AssocConstraintType:
	case kAngle2AssocConstraintType:
	case kAngle3AssocConstraintType:
		namePrefix = ACRX_T("ang");
		break;

	case kRadiusAssocConstraintType:
		namePrefix = ACRX_T("rad");
		break;

	case kDiameterAssocConstraintType:
		namePrefix = ACRX_T("dia");
		break;

	default:
		assert(!ACRX_T("Unsupported dimensional constraint type"));
		namePrefix = ACRX_T("error");
		break;
	}

	AcString dimensionText;
	double	 measurement = 0.0;
	bool	 isAngular = false;
	{
		AcDbObjectPointer<AcDbDimension> pDim(dimId, kForRead);
		if (pDim.openStatus() != eOk)
			return err;

		if (pDim->isConstraintObject())
		{
			LPTSTR pDimText = pDim->dimensionText();
			if (pDimText)
			{
				if (*pDimText)
					dimensionText = pDimText;
				free(pDimText);
			}
		}

		err = pDim->measurement(measurement);
		isAngular = pDim->isKindOf(AcDb2LineAngularDimension::desc()) ||
					pDim->isKindOf(AcDb3PointAngularDimension::desc());
	}

	// Extract the dimension name and expression given the entity text
	err = AcDbAssocDimDependencyBodyBase::getNameAndExpressionFromEntityText(
			dimensionText, true, measurement, isAngular, dfltName, dfltExpression);
	assert(!dfltExpression.isEmpty());

	//default name should be empty if it's not a dimconstraint, need to generate a unique name from prefix
	if (dfltName.isEmpty())
	{
		const AcDbObjectId networkId = getAssocNetwork(true, NULL);
		AcDbObjectPointer<AcDbAssocNetwork> pNetwork(networkId, kForRead);
		if (pNetwork.openStatus() != eOk)
		{
			assert(!ACRX_T("Cannot open an Assoc Network"));
			return eInvalidInput;
		}

		AcDbAssocNetworkIterator iter(pNetwork);
		AcArray<int> suffixArray;
		while (iter.moveNext())
		{
			AcDbObjectPointer<AcDbAssocVariable> pVar(iter.current(), kForRead, true);
			if (pVar.openStatus() != eOk)
				continue;

			const AcString varName = pVar->name();
			if (varName.matchNoCase(namePrefix) == namePrefix.length())
			{
				const AcString number = varName.substr(namePrefix.length(), -1);
				int val = number.asDeci(10);
				if (val >= 0)
				{
					suffixArray.append(val);
				}

			}
		}

		int index = 1;
		while (suffixArray.find(index) != -1)
		{
			if (index < 0)
			{
				assert(!"Interger overflow");
				return eInvalidInput;
			}
			++index;
		}

		dfltName = namePrefix;
		const int kBufSize = 20;
		char buf[kBufSize];
		const errno_t err = _itoa_s(index, buf, kBufSize, 10);
		assert(err == 0);
		dfltName += buf;
	}

	name		 = dfltName;
	expression	 = dfltExpression;

	return err;
}

/*
* Name: addNewVariableToAssocNetwork
*
* Parameters:
* [INPUT]const AcString &name: The name of the new variable
* [INPUT]const AcString &expression: The expression of the new variable
* [INPUT]const AcDbObjectId &networkId: The network to which the variable and dependency will be added
* [OUTPUT]AcDbObjectId &varId: The ObjectID of the variable
* [OUTPUT]AcDbObjectId &varDepId: The ObjectID of the dependency
*
* Return: Error status
*/
Acad::ErrorStatus
SampleUtils::addNewVariableToAssocNetwork(const AcString &name,
														const AcString &expression,
														const AcDbObjectId &networkId,
														AcDbObjectId &varId,
														AcDbObjectId &varDepId)
{
	Acad::ErrorStatus eStatus = eOk;
	varId = AcDbObjectId::kNull;
	varDepId = AcDbObjectId::kNull;

	AcDbObjectPointer<AcDbAssocNetwork> pNetwork(networkId, kForWrite);
	ASSERT(pNetwork.openStatus() == eOk);

	{
		AcDbAssocVariable* const pVar = new AcDbAssocVariable();
		eStatus = pNetwork->database()->addAcDbObject(varId, pVar);
		ASSERT(eStatus == eOk);
		pVar->close();

		eStatus = pNetwork->addAction(varId, true);
		ASSERT(eStatus == eOk);
	}

	AcDbObjectPointer<AcDbAssocVariable> pVar (varId, kForWrite);
	eStatus = pVar->setName(name, true);
	
	AcDbEvalVariant val(2.0);
	pVar->setValue(val);

	eStatus = pVar->setExpression(expression, EXPRESSION_EVALUATOR_DEFAULT, true, true);

	AcString errorMessage(_T("Expression evaluation is failed"));
	AcDbEvalVariant evaluatedExpressionValue;
	eStatus = pVar->evaluateExpression(evaluatedExpressionValue, errorMessage);
	if (eStatus == eOk)
		eStatus = pVar->setValue(evaluatedExpressionValue);
	
	AcDbAssocValueDependency* const pValDep = new AcDbAssocValueDependency();
	networkId.database()->addAcDbObject(varDepId, pValDep);
	pValDep->attachToObject(varId);
	pValDep->close();

	ASSERT(!varId.isNull() && !varDepId.isNull());
	return eStatus;
}

/*
* Name: createDistanceConstraint
*
* Parameters:
* [INPUT]AcDistanceConstraint::DirectionType directionType: The direction type of constraint. For aligned dimensions it should be kNotDirected
* [INPUT]AcDbObjectId varDepId: The objectid of variable dependency object
* [INPUT]AcDbObjectId dimDepId: The objectid of dimensional dependency object
* [INPUT]AcConstrainedGeometry*& pConsGeom: The constrained subentity geometry of the entity being constrained
* [INPUT]AcDbFullSubentPath& subEntPath: The subentity path of the subentity of the entity being constrained
* [INPUT]const AcGeVector3d *pFixedDirection: The direction of the dimension if the direction type is fixed
*
* Return: Error status
*/
Acad::ErrorStatus SampleUtils::createDistanceConstraint(AcDistanceConstraint::DirectionType directionType,
		AcDbObjectId varDepId,
		AcDbObjectId dimDepId,
		AcConstrainedGeometry*& pConsGeom,
		AcDbFullSubentPath& subEntPath,
		const AcGeVector3d *pFixedDirection)
{
	ErrorStatus eStatus = Acad::eNotImplementedYet;
	
	AcDbObjectPointer<AcDbAssoc2dConstraintGroup> p2dConstrGrp(getConstraintGroup(NULL, true), kForWrite);
	eStatus = p2dConstrGrp.openStatus();
	
	
		AcArray<AcConstrainedImplicitPoint*> apImplicitPoints;
		AcConstrainedCurve* pEntToConstrained = NULL;

	
		if (pConsGeom->isA() == AcConstrainedBoundedLine::desc())
			pEntToConstrained = AcConstrainedLine::cast(pConsGeom);
		else
			return eInvalidInput;

		AcConstrainedGeometry* pConsPoint[2] = {NULL, NULL,};
		pEntToConstrained->getConstrainedImplicitPoints(apImplicitPoints);

		for (int i = 0; i < apImplicitPoints.length(); i++)
		{
			if (apImplicitPoints[i]->pointType() == AcConstrainedImplicitPoint::kStartImplicit)
				pConsPoint[0] = apImplicitPoints[i];
			else if (apImplicitPoints[i]->pointType() == AcConstrainedImplicitPoint::kEndImplicit)
				pConsPoint[1] = apImplicitPoints[i];
		}
		
	
	eStatus = p2dConstrGrp->addDistanceConstraint(pConsPoint[0],
												  pConsPoint[1],
												  directionType,
												  varDepId,
												  dimDepId,
												  pFixedDirection);
	return eStatus;
}