﻿
'This sample demonstrates the Overrule Protocol with a simple example of turning a line into a thermometer.
' A more complete version will be posted in the future.


#Region "HelperClass"



'Global helper class (singleton). Contains central definitions of some global constants, and a few helper functions
Public Class HelperClass
    Const mExtDictName As String = "SGP_MyDict" 'Defines Dictionary name for the Extension Dictionary demo
    Const mXRecName As String = "SGP_MyData" 'Defines Dictionary name for the Extension Dictionary demo
    Const mXRecName2 As String = "SGP_MyLink"
    Const mSize As Integer = 30 'Universal scaling constant - so I don't have to edit every calculation if I want the thermometer thicker or thinner


    Private Shared mMe As HelperClass


    'Name of our dictionary in extension dictionary
    Public ReadOnly Property DictionaryName()
        Get
            Return mExtDictName
        End Get
    End Property

    'Name of our XRecord
    Public ReadOnly Property XRecordName()
        Get
            Return mXRecName
        End Get
    End Property


    'Name of our 2nd XRecord (where we store hardpointer to linked entity)
    Public ReadOnly Property XRecordName2()
        Get
            Return mXRecName2
        End Get
    End Property



    'Protected constructor - to enforce singleton behavior
    Protected Sub New()

    End Sub

    'static function to retrieve one and only instance of singleton
    Shared ReadOnly Property GetSingleton()
        Get
            If mMe Is Nothing Then
                mMe = New HelperClass
            End If
            Return mMe
        End Get
    End Property

    'Retrieve data (as resbuf) from or Xrecord.
    'Returns null object if there's a problem
    Public Function GetXRecordData(ByVal obj As DBObject, ByVal name As String) As ResultBuffer

        Dim xRec As Xrecord = Nothing
        Dim id As ObjectId = obj.ExtensionDictionary

        'Make sure we have an ext dict befoore proceeding
        If id.IsValid Then

            'Retrieve data using a transaction
            Dim db As Database = obj.Database

            '*** IMPORTANT ***
            'We're using OpenCloseTransaction rather than usual transaction to prevent problems caused by using Transactions in worldDraw. 
            'Alternative is to use ObjectID.Open to open objects outside transaction (which is really what the OpenCloseTransaction is doing).
            'When using ObjectId.Open, remember to either explicitly Close the object, or use a Using statement to force the object to go out of scope and be automatically Closed.
            '***************** 
            Using tr As OpenCloseTransaction = db.TransactionManager.StartOpenCloseTransaction

                Dim extDict As DBDictionary = tr.GetObject(id, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForRead, False)
                If extDict.Contains(DictionaryName) Then
                    Dim dictId As ObjectId = extDict.GetAt(DictionaryName)
                    Dim myDict As DBDictionary = tr.GetObject(dictId, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForRead, False)
                    If myDict.Contains(name) Then
                        xRec = tr.GetObject(myDict.GetAt(name), Autodesk.AutoCAD.DatabaseServices.OpenMode.ForRead, False)
                    End If
                End If
            End Using
        End If
        If xRec Is Nothing Then
            Return Nothing
        Else
            Return xRec.Data
        End If
    End Function


    'Modifies data in our XRecord.
    '(creates our dictionary and XRecoird if it doesn't already exist)
    Public Sub SetXRecordData(ByVal obj As DBObject, ByVal myData As ResultBuffer, ByVal name As String)

        Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database

        'Use OpenCloseTransaction in case this function gets called somewhere transactions (and associated UNDO recording) are a bad idea
        '(see also GetXRecordData function)
        Using tr As OpenCloseTransaction = db.TransactionManager.StartOpenCloseTransaction

            Dim myDict As DBDictionary
            Dim xRec As Xrecord = Nothing

            Dim id As ObjectId = obj.ExtensionDictionary

            If id = ObjectId.Null Then
                obj.CreateExtensionDictionary()
                id = obj.ExtensionDictionary
            End If


            Dim extDict As DBDictionary = tr.GetObject(id, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForWrite, False)

            If extDict.Contains(DictionaryName) Then
                Dim dictId As ObjectId = extDict.GetAt(DictionaryName)
                myDict = tr.GetObject(dictId, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForWrite, False)
            Else
                myDict = New DBDictionary
                extDict.SetAt(DictionaryName, myDict)
                tr.AddNewlyCreatedDBObject(myDict, True)
            End If


            If myDict.Contains(name) Then
                xRec = tr.GetObject(myDict.GetAt(name), Autodesk.AutoCAD.DatabaseServices.OpenMode.ForWrite, False)
            Else
                xRec = New Xrecord
                myDict.SetAt(name, xRec)
                tr.AddNewlyCreatedDBObject(xRec, True)
            End If

            Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("Adding Xdata to Object with ID = " & obj.ObjectId.ToString & vbLf)
            xRec.Data = myData

            'Easy way to ensure Data in extension dictionary (and sub-dictionary) is copied when deepcloning.
            extDict.TreatElementsAsHard = True
            myDict.TreatElementsAsHard = True

            tr.Commit()
        End Using
    End Sub


    Public Sub ThermometerExplodeFragments(ByVal theLine As Line, ByVal entitySet As DBObjectCollection)

        'Read Xrecord values to use in drawing temperature related graphics (mercury and levels)
        Dim resbuf As ResultBuffer = HelperClass.GetSingleton.GetXRecordData(theLine, XRecordName)

        'If there's no data, then we just exit.
        If resbuf Is Nothing Then
            Exit Sub
        End If

        'Add 'copy' of original line to explode set
        Dim ent As Entity = New Line(theLine.StartPoint, theLine.EndPoint)
        ent.SetDatabaseDefaults()
        entitySet.Add(ent)

        'Extract data from resbuf to use for thermometer markers
        Dim myText As String = resbuf.AsArray(0).Value 'Room name
        Dim lowerTemp As Integer = resbuf.AsArray(1).Value 'Min temp
        Dim upperTemp As Integer = resbuf.AsArray(2).Value 'max temp
        Dim curTemp As Integer = resbuf.AsArray(3).Value 'Current temp

        Dim curPos As Double = curTemp / 100
        Dim perpVec As Vector3d = (theLine.EndPoint - theLine.StartPoint).CrossProduct(theLine.Normal).GetNormal

        Dim startParam As Double = theLine.GetParameterAtPoint(theLine.StartPoint)
        Dim endParam As Double = theLine.GetParameterAtPoint(theLine.EndPoint)


        Dim posParam As Double

        'Draw thermometer body

        'right body edge
        ent = New Line(theLine.StartPoint + perpVec * theLine.Length * 2.5 / mSize, theLine.EndPoint + perpVec * theLine.Length * 2.5 / mSize)
        ent.SetDatabaseDefaults()
        entitySet.Add(ent)

        'left body edge
        ent = New Line(theLine.EndPoint - perpVec * theLine.Length * 2.5 / mSize, theLine.StartPoint - perpVec * theLine.Length * 2.5 / mSize)
        ent.SetDatabaseDefaults()
        entitySet.Add(ent)

        'top body edge

        Dim gArc As New CircularArc3d(theLine.EndPoint + perpVec * theLine.Length * 2.5 / mSize, theLine.EndPoint + (theLine.EndPoint - theLine.StartPoint) * 2.5 / mSize, theLine.EndPoint - perpVec * theLine.Length * 2.5 / mSize)
        Dim refAngle As Double = Vector3d.XAxis.GetAngleTo(gArc.ReferenceVector, gArc.Normal)

        ent = New Arc(gArc.Center, theLine.Normal, gArc.Radius, refAngle + gArc.StartAngle, refAngle + gArc.EndAngle)
        ent.SetDatabaseDefaults()
        entitySet.Add(ent)


        'bottom body edge
        Dim theta As Double = Math.PI / 6
        Dim rad As Double = (theLine.Length * 2.5 / mSize) / Math.Sin(theta)
        Dim a As Double = (theLine.Length * 2.5 / mSize) / Math.Tan(theta)
        Dim bowlCenter As Point3d = theLine.StartPoint + (theLine.StartPoint - theLine.EndPoint).GetNormal * a

        gArc = New CircularArc3d(theLine.StartPoint - perpVec * theLine.Length * 2.5 / mSize, _
                                theLine.StartPoint + (theLine.StartPoint - theLine.EndPoint).GetNormal * (rad + a), _
                                 theLine.StartPoint + perpVec * theLine.Length * 2.5 / mSize)
        refAngle = Vector3d.XAxis.GetAngleTo(gArc.ReferenceVector, gArc.Normal)

        ent = New Arc(gArc.Center, gArc.Normal, gArc.Radius, refAngle + gArc.StartAngle, refAngle + gArc.EndAngle)
        ent.SetDatabaseDefaults()
        entitySet.Add(ent)

        'Draw upper temperature marker
        posParam = startParam + (endParam - startParam) * (upperTemp / 100)
        ent = New Line(theLine.GetPointAtParameter(posParam) - perpVec * theLine.Length * 3 / mSize, theLine.GetPointAtParameter(posParam) + perpVec * theLine.Length * 3 / mSize)
        ent.SetDatabaseDefaults()
        entitySet.Add(ent)

        'Draw text
        Dim txt As New DBText
        txt.SetDatabaseDefaults()
        txt.Position = theLine.GetPointAtParameter(posParam) + perpVec * theLine.Length * 4 / mSize
        txt.TextString = "Max. Temp = " & upperTemp.ToString
        entitySet.Add(txt)


        'Draw lower temperature marker
        posParam = startParam + (endParam - startParam) * (lowerTemp / 100)
        entitySet.Add(New Line(theLine.GetPointAtParameter(posParam) - perpVec * theLine.Length * 3 / mSize, _
        theLine.GetPointAtParameter(posParam) + perpVec * theLine.Length * 3 / mSize))

        'Draw text
        txt = New DBText
        txt.SetDatabaseDefaults()
        txt.Position = theLine.GetPointAtParameter(posParam) + perpVec * theLine.Length * 4 / mSize
        txt.TextString = "Min. Temp = " & lowerTemp.ToString
        entitySet.Add(txt)




    End Sub


End Class


#End Region



#Region "Simple GripOverrule"


'Grip overrule to add our custom grips to the line
Public Class MyGripOverrule
    Inherits GripOverrule


    Public Overrides Sub GetGripPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal gripPoints As Autodesk.AutoCAD.Geometry.Point3dCollection, ByVal snapModes As Autodesk.AutoCAD.Geometry.IntegerCollection, ByVal geometryIds As Autodesk.AutoCAD.Geometry.IntegerCollection)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GripOverrule.GetGripPoints 1 called" & vbLf)

        MyBase.GetGripPoints(entity, gripPoints, snapModes, geometryIds)
    End Sub


    Public Overrides Sub GetStretchPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal stretchPoints As Autodesk.AutoCAD.Geometry.Point3dCollection)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GripOverrule.GetStretchPoints called" & vbLf)
        MyBase.GetStretchPoints(entity, stretchPoints)
    End Sub


    Public Overrides Sub MoveGripPointsAt(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal indices As Autodesk.AutoCAD.Geometry.IntegerCollection, ByVal offset As Autodesk.AutoCAD.Geometry.Vector3d)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GripOverrule.MoveGripPointsAt 1 called" & vbLf)
        MyBase.MoveGripPointsAt(entity, indices, offset)
    End Sub


    Public Overrides Sub MoveStretchPointsAt(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal indices As Autodesk.AutoCAD.Geometry.IntegerCollection, ByVal offset As Autodesk.AutoCAD.Geometry.Vector3d)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GripOverrule.MoveStretchPointsAt called" & vbLf)
        MyBase.MoveStretchPointsAt(entity, indices, offset)
    End Sub

    Public Overrides Sub OnGripStatusChanged(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal status As Autodesk.AutoCAD.DatabaseServices.GripStatus)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GripOverrule.OnGripStatusChanged called" & vbLf)
        MyBase.OnGripStatusChanged(entity, status)
    End Sub

    'Our custom grip class
    '(Could have derived one class for each grip, but we'll use member dara (Ordinal property) to distinguis grips instead)
    Public Class MyGrip
        Inherits GripData
        Private mGripNum As Integer

        Public Property Ordinal() As Integer
            Get
                Return mGripNum
            End Get
            Set(ByVal value As Integer)
                mGripNum = value
            End Set
        End Property

        'Call this to tell the grip to move itself
        Public Sub Move(ByVal vec As Vector3d)
            GripPoint = GripPoint + vec
        End Sub

        'Grip draws itself
        Public Overrides Function ViewportDraw(ByVal worldDraw As Autodesk.AutoCAD.GraphicsInterface.ViewportDraw, ByVal entityId As Autodesk.AutoCAD.DatabaseServices.ObjectId, ByVal type As Autodesk.AutoCAD.DatabaseServices.GripData.DrawType, ByVal imageGripPoint As Autodesk.AutoCAD.Geometry.Point3d?, ByVal gripSizeInPixels As Integer) As Boolean
            Dim unit As Point2d = worldDraw.Viewport.GetNumPixelsInUnitSquare(GripPoint)
            worldDraw.Geometry.Circle(GripPoint, 1.5 * gripSizeInPixels / unit.X, worldDraw.Viewport.ViewDirection)
            Return True
        End Function
    End Class


    'Array to hold our 3 grips
    Dim mGripData(2) As GripData



    Public Overrides Sub GetGripPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal grips As Autodesk.AutoCAD.DatabaseServices.GripDataCollection, ByVal curViewUnitSize As Double, ByVal gripSize As Integer, ByVal curViewDir As Autodesk.AutoCAD.Geometry.Vector3d, ByVal bitFlags As Autodesk.AutoCAD.DatabaseServices.GetGripPointsFlags)

        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GripOverrule.GetGripPoints 2 called" & vbLf)

        Dim rb As ResultBuffer = HelperClass.GetSingleton.GetXRecordData(entity, HelperClass.GetSingleton.XRecordName)

        'Only draw custom grips if we have valid data and entity is a line
        If (Not rb Is Nothing) And (TypeOf entity Is Line) Then
            Dim myLine As Line = entity

            'Set grip positions to represent temperatures (we're using Celsius)

            'min temperature
            Dim temp As Integer = rb.AsArray(1).Value
            Dim pos As Double = myLine.StartParam + (temp / 100) * (myLine.EndParam - myLine.StartParam)
            Dim pt As Point3d = myLine.GetPointAtParameter(pos)
            Dim grip As New MyGrip
            grip.Ordinal = 0
            grip.GripPoint = pt
            mGripData(0) = grip

            'max temperature
            temp = rb.AsArray(2).Value
            pos = myLine.StartParam + (temp / 100) * (myLine.EndParam - myLine.StartParam)
            pt = myLine.GetPointAtParameter(pos)
            grip = New MyGrip
            grip.Ordinal = 1
            grip.GripPoint = pt
            mGripData(1) = grip

            'current temperature
            temp = rb.AsArray(3).Value
            pos = myLine.StartParam + (temp / 100) * (myLine.EndParam - myLine.StartParam)
            pt = myLine.GetPointAtParameter(pos)
            grip = New MyGrip
            grip.Ordinal = 2
            grip.GripPoint = pt
            mGripData(2) = grip

            'Add our grips to the list
            For Each g As MyGrip In mGripData
                grips.Add(g)
            Next

        End If
        'Get the standard line grip points as well
        MyBase.GetGripPoints(entity, grips, curViewUnitSize, gripSize, curViewDir, bitFlags)

    End Sub



    Public Overrides Sub MoveGripPointsAt(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal grips As Autodesk.AutoCAD.DatabaseServices.GripDataCollection, ByVal offset As Autodesk.AutoCAD.Geometry.Vector3d, ByVal bitFlags As Autodesk.AutoCAD.DatabaseServices.MoveGripPointsFlags)

        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GripOverrule.MoveGripPointsAt 2 called" & vbLf)

        'We only take  action when we get this call on a database resident entity
        'Dragging operation makes shallow clone of line, and setting clomeMeForDragging to false is generally a bad idea.
        '(If you do set clone me for dragging to false, then don't call bae class overriden methods).
        If entity.Id.IsValid Then

            'Get temperature data
            Dim rb As ResultBuffer = HelperClass.GetSingleton.GetXRecordData(entity, HelperClass.GetSingleton.XRecordName)

            'Only draw custom grips if we have valid data and entity is a line
            If (Not rb Is Nothing) And (TypeOf entity Is Line) Then


                'Cast to a Line so we can access properties
                Dim myLine As Line = entity

                Dim lineDir As Vector3d = (myLine.EndPoint - myLine.StartPoint)
                lineDir = lineDir.GetNormal 'Direction of Line
                Dim offsetDist As Double = lineDir.DotProduct(offset) 'Component of mouse translation along like

                'Iterate through list of all grips being moved
                For Each g As GripData In grips
                    If TypeOf g Is MyGrip Then
                        Dim grip As MyGrip = g 'Cast to our grip type

                        'Make sure offset never takes grip beyond either end of line
                        If offsetDist >= 0 Then
                            If offsetDist > (myLine.EndPoint - grip.GripPoint).Length Then
                                offsetDist = (myLine.EndPoint - grip.GripPoint).Length
                            End If
                        Else
                            If -offsetDist > (myLine.StartPoint - grip.GripPoint).Length Then
                                offsetDist = -(myLine.StartPoint - grip.GripPoint).Length
                            End If
                        End If
                        lineDir = lineDir * offsetDist

                        'retrieve stored data and edit the changed value
                        Dim val1 As String = rb.AsArray(0).Value
                        Dim intVal(2) As Integer
                        intVal(0) = rb.AsArray(1).Value 'min
                        intVal(1) = rb.AsArray(2).Value 'max
                        intVal(2) = rb.AsArray(3).Value 'current

                        'Tell grip to move itself long the line
                        grip.Move(lineDir)

                        'Calculate new temperature from grip position along the line
                        Dim newParam As Double = myLine.GetParameterAtPoint(grip.GripPoint)
                        Dim newTemp As Integer = 100 * (newParam - myLine.StartParam) / (myLine.EndParam - myLine.StartParam)

                        'Don't let min temp value rise above max temp
                        'And don't let max temp go below min temp
                        If grip.Ordinal = 0 Then
                            If newTemp < intVal(1) Then
                                intVal(0) = newTemp
                            Else
                                intVal(0) = intVal(1) - 1
                            End If
                        ElseIf grip.Ordinal = 1 Then
                            If newTemp > intVal(0) Then
                                intVal(1) = newTemp
                            Else
                                intVal(1) = intVal(0) + 1
                            End If
                        Else
                            intVal(2) = newTemp
                        End If

                        'Create new resbuf with new data and put back in Xrecord
                        Dim newRb As ResultBuffer = New ResultBuffer(New TypedValue(DxfCode.Text, val1), _
                                              New TypedValue(DxfCode.Int32, intVal(0)), _
                                              New TypedValue(DxfCode.Int32, intVal(1)), _
                                              New TypedValue(DxfCode.Int32, intVal(2)))
                        HelperClass.GetSingleton.SetXRecordData(myLine, newRb, HelperClass.GetSingleton.XRecordName)
                    End If
                Next
            End If
        End If

        'Remove our grips from the list befroe calling base class function
        '(Doesn't seem to like my grips)
        For i As Integer = grips.Count - 1 To 0 Step -1
            If TypeOf grips(i) Is MyGrip Then
                grips.Remove(grips(i))
            End If
        Next
        'If any grips left, then we call base class function
        If grips.Count > 0 Then
            MyBase.MoveGripPointsAt(entity, grips, offset, bitFlags)
        End If

    End Sub



End Class


#End Region




#Region "Simple DrawableOverrule "

'This overrule adds our custom graphhics to the Line
'We're going to turn our Line into a Thermometer
Public Class MyDrawOverrule
    Inherits DrawableOverrule

    Const mSize As Integer = 30 'Universal scaling constant - so I don't have to edit every calculation if I want the thermometer thicker or thinner


    Public Overrides Function SetAttributes(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal traits As Autodesk.AutoCAD.GraphicsInterface.DrawableTraits) As Integer
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("drawableOverrule.SetAttributes called" & vbLf)
        Return MyBase.SetAttributes(drawable, traits)
    End Function


    Public Overrides Sub ViewportDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal vd As Autodesk.AutoCAD.GraphicsInterface.ViewportDraw)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("drawableOverrule.ViewportDraw called" & vbLf)
        MyBase.ViewportDraw(drawable, vd)
    End Sub


    Public Overrides Function ViewportDrawLogicalFlags(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal vd As Autodesk.AutoCAD.GraphicsInterface.ViewportDraw) As Integer
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("drawableOverrule.ViewportDrawLogicalFlags called" & vbLf)
        Return MyBase.ViewportDrawLogicalFlags(drawable, vd)
    End Function



    'This is the function that gets called to add/replace an entity's WorldDraw graphics
    Public Overrides Function WorldDraw(ByVal drawable As Autodesk.AutoCAD.GraphicsInterface.Drawable, ByVal wd As Autodesk.AutoCAD.GraphicsInterface.WorldDraw) As Boolean

        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("drawableOverrule.WorldDraw called" & vbLf)

        'Is it a line? (It should be)
        If Not TypeOf (drawable) Is Line Then Return MyBase.WorldDraw(drawable, wd)
        Dim myLine As Line = drawable
        Dim pts As New Point3dCollection

        'Read Xrecord values to use in drawing temperature related graphics (mercury and levels)
        Dim resbuf As ResultBuffer = HelperClass.GetSingleton.GetXRecordData(myLine, HelperClass.GetSingleton.XRecordName)

        'If there's no data, then we just call the base class's WorldDraw function
        If Not resbuf Is Nothing Then

            Dim myText As String = resbuf.AsArray(0).Value 'Room name
            Dim lowerTemp As Integer = resbuf.AsArray(1).Value 'Min temp
            Dim upperTemp As Integer = resbuf.AsArray(2).Value 'max temp
            Dim curTemp As Integer = resbuf.AsArray(3).Value 'Current temp

            Dim curPos As Double = curTemp / 100
            Dim perpVec As Vector3d = (myLine.EndPoint - myLine.StartPoint).CrossProduct(myLine.Normal).GetNormal

            Dim startParam As Double = myLine.GetParameterAtPoint(myLine.StartPoint)
            Dim endParam As Double = myLine.GetParameterAtPoint(myLine.EndPoint)

            Dim oldColIndex = wd.SubEntityTraits.Color
            Dim oldFillType As FillType = wd.SubEntityTraits.FillType

            Dim posParam As Double
            Dim gsMarker As IntPtr

            'Draw thermometer body
            wd.SubEntityTraits.FillType = FillType.FillNever

            'right body edge
            pts.Clear()
            pts.Add(myLine.StartPoint + perpVec * myLine.Length * 2.5 / mSize)
            pts.Add(myLine.EndPoint + perpVec * myLine.Length * 2.5 / mSize)
            gsMarker = 1
            wd.Geometry.Polyline(pts, myLine.Normal, gsMarker)

            'left body edge
            pts.Clear()
            pts.Add(myLine.EndPoint - perpVec * myLine.Length * 2.5 / mSize)
            pts.Add(myLine.StartPoint - perpVec * myLine.Length * 2.5 / mSize)
            gsMarker = 2
            wd.Geometry.Polyline(pts, myLine.Normal, gsMarker)

            'top body edge
            wd.Geometry.CircularArc(myLine.EndPoint - perpVec * myLine.Length * 2.5 / mSize, myLine.EndPoint + (myLine.EndPoint - myLine.StartPoint) * 2.5 / mSize, myLine.EndPoint + perpVec * myLine.Length * 2.5 / mSize, ArcType.ArcSimple)

            'bottom body edge
            Dim theta As Double = Math.PI / 6
            Dim rad As Double = (myLine.Length * 2.5 / mSize) / Math.Sin(theta)
            Dim a As Double = (myLine.Length * 2.5 / mSize) / Math.Tan(theta)
            Dim bowlCenter As Point3d = myLine.StartPoint + (myLine.StartPoint - myLine.EndPoint).GetNormal * a
            wd.Geometry.CircularArc(myLine.StartPoint + perpVec * myLine.Length * 2.5 / mSize, _
                                    myLine.StartPoint + (myLine.StartPoint - myLine.EndPoint).GetNormal * (rad + a), _
                                    myLine.StartPoint - perpVec * myLine.Length * 2.5 / mSize, _
                                    ArcType.ArcSimple)

            'Draw upper temperature marker (in red)
            wd.SubEntityTraits.Color = 1
            posParam = startParam + (endParam - startParam) * (upperTemp / 100)
            pts.Clear()
            pts.Add(myLine.GetPointAtParameter(posParam) - perpVec * myLine.Length * 3 / mSize)
            pts.Add(myLine.GetPointAtParameter(posParam) + perpVec * myLine.Length * 3 / mSize)
            gsMarker = 3
            wd.Geometry.Polyline(pts, myLine.Normal, gsMarker)

            'Default TextStyle constructor uses current (default) textstyle
            Dim txtStyle As New TextStyle

            wd.Geometry.Text(myLine.GetPointAtParameter(posParam) + perpVec * myLine.Length * 4 / mSize, _
                             myLine.Normal, _
                             perpVec, _
                             "Max. Temp = " & upperTemp.ToString, _
                             True, _
                             txtStyle)


            'Draw lower temperature marker (in blue)
            wd.SubEntityTraits.Color = 5
            posParam = startParam + (endParam - startParam) * (lowerTemp / 100)
            pts.Clear()
            pts.Add(myLine.GetPointAtParameter(posParam) - perpVec * myLine.Length * 3 / mSize)
            pts.Add(myLine.GetPointAtParameter(posParam) + perpVec * myLine.Length * 3 / mSize)
            gsMarker = 5
            wd.Geometry.Polyline(pts, myLine.Normal, gsMarker)
            wd.Geometry.Text(myLine.GetPointAtParameter(posParam) + perpVec * myLine.Length * 4 / mSize, _
                             myLine.Normal, _
                             perpVec, _
                             "Min. Temp = " & lowerTemp.ToString, _
                             True, _
                             txtStyle)

            'Draw current temperature marker in different color depending on position w.r.t. min and max temps
            Dim colIndex As Integer
            If curTemp <= lowerTemp Then
                colIndex = 5 'Blue
            ElseIf curTemp >= upperTemp Then
                colIndex = 1 'Red
            Else
                colIndex = 94 'Dark green
            End If

            'Draw current Temperature marker
            wd.SubEntityTraits.Color = colIndex

            posParam = startParam + (endParam - startParam) * (curTemp / 100)
            pts.Clear()
            pts.Add(myLine.GetPointAtParameter(posParam) - perpVec * myLine.Length * 3 / mSize)
            pts.Add(myLine.GetPointAtParameter(posParam) + perpVec * myLine.Length * 3 / mSize)
            gsMarker = 6
            wd.Geometry.Polyline(pts, myLine.Normal, gsMarker) '(myLine.GetPointAtParameter(posParam), myLine.Length / mSize, myLine.Normal)
            'wd.Geometry.Circle(myLine.GetPointAtParameter(posParam), myLine.Length / 30, myLine.Normal)
            wd.Geometry.Text(myLine.GetPointAtParameter(posParam) + perpVec * myLine.Length * 4 / mSize, _
                 myLine.Normal, _
                 perpVec, _
                 myText & " Temp = " & curTemp.ToString, _
                 True, _
                 txtStyle)


            'We want to draw filled primitives (polygon and circle) to represent the mercury in the thermometer
            wd.SubEntityTraits.FillType = FillType.FillAlways

            'drawable mercury - line first, then bowl
            pts.Clear()
            Dim offset As Vector3d = perpVec * myLine.Length / mSize
            Dim pt1 As Point3d = myLine.StartPoint + offset
            pts.Add(bowlCenter + offset)
            pts.Add(bowlCenter - offset)
            pts.Add(myLine.GetPointAtParameter(posParam) - offset)
            pts.Add(myLine.GetPointAtParameter(posParam) + offset)
            wd.Geometry.Polygon(pts)

            'mercury bowl
            theta = Math.PI / 6
            rad = 1.5 * (offset.Length) / Math.Sin(theta)
            a = (offset.Length) / Math.Tan(theta)
            wd.Geometry.Circle(bowlCenter, rad, myLine.Normal)

            'Set old subentitytrait values, then call overriden class worlddraw fn
            wd.SubEntityTraits.FillType = oldFillType
            wd.SubEntityTraits.Color = oldColIndex

        End If

        Return MyBase.WorldDraw(drawable, wd)

    End Function


End Class

#End Region

#Region "Transform Overrule"

Public Class MyTransformOverrule
    Inherits TransformOverrule

    Const mSize As Integer = 30 'Universal scaling constant - so I don't have to edit every calculation if I want the thermometer thicker or thinner

    'Note - There is currently a bug in TransformOverrule.Explode:
    'If you call Explode method, and don't append objects to the database, then AutoCAD crashes.
    '(The EXPLODE command works fine with the overrule in place). 
    Public Overrides Sub Explode(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal entitySet As Autodesk.AutoCAD.DatabaseServices.DBObjectCollection)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("TransformOverrule.Explode called" & vbLf)

        'Is it a line? (It should be)
        ' But if it isn't, then we call base class explode function
        If Not TypeOf (entity) Is Line Then
            MyBase.Explode(entity, entitySet)
            Exit Sub
        End If

        Dim myLine As Line = entity

        Dim explodedEntities As DBObjectCollection = HelperClass.GetSingleton.ThermometerExplodeFragments(myLine)

        For Each ent As Entity In explodedEntities
            entitySet.Add(ent)
        Next

 
    End Sub



    Public Overrides Sub TransformBy(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal transform As Autodesk.AutoCAD.Geometry.Matrix3d)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("TransformOverrule.Transformby called" & vbLf)
        MyBase.TransformBy(entity, transform)
    End Sub

    Public Overrides Function CloneMeForDragging(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity) As Boolean
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("TransformOverrule.CloneMeForDragging called" & vbLf)
        Return MyBase.CloneMeForDragging(entity)
    End Function

    Public Overrides Function HideMeForDragging(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity) As Boolean
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("TransformOverrule.HideMeForDragging called" & vbLf)
        Return MyBase.HideMeForDragging(entity)
    End Function


    Public Overrides Function GetTransformedCopy(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal transform As Autodesk.AutoCAD.Geometry.Matrix3d) As Autodesk.AutoCAD.DatabaseServices.Entity
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("TransformOverrule.GetTransformedCopy called" & vbLf)
        Return MyBase.GetTransformedCopy(entity, transform)
    End Function

End Class
#End Region


#Region "Properties Overrule"


Public Class MyPropertiesOverrule
    Inherits PropertiesOverrule

    Public Overrides Function GetClassID(ByVal entity As Autodesk.AutoCAD.DatabaseServices.DBObject) As System.Guid
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("PropertiesOverrule.GetClassID called" & vbLf)
        Return MyBase.GetClassID(entity)
    End Function

    Public Overrides Sub List(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("PropertiesOverrule.List called" & vbLf)

        If Not TypeOf entity Is Line Then
            MyBase.List(entity)
            Exit Sub
        End If

        'Get data from XRecord so we can LIST it.
        Dim resbuf As ResultBuffer = HelperClass.GetSingleton.GetXRecordData(entity, HelperClass.GetSingleton.XRecordName)

        'If there's no data, then we just call the base class's WorldDraw function
        If resbuf Is Nothing Then
            MyBase.List(entity)
            Exit Sub
        End If

        Dim myText As String = resbuf.AsArray(0).Value 'City name
        Dim lowerTemp As Integer = resbuf.AsArray(1).Value 'Min temp
        Dim upperTemp As Integer = resbuf.AsArray(2).Value 'max temp
        Dim curTemp As Integer = resbuf.AsArray(3).Value 'Current temp

        'LIST entity data
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage(vbLf & "-------------------------------------" & vbLf)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage(vbLf & "THERMOMETER" & vbLf)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("Minimum Temperature = " & lowerTemp.ToString & vbLf)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("Maximum Temperature = " & upperTemp.ToString & vbLf)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage(myText & " Temperature = " & curTemp.ToString & vbLf)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage(vbLf & "-------------------------------------" & vbLf)

    End Sub

End Class

#End Region



#Region "Object Overrule"

Public Class MyObjectOverrule
    Inherits ObjectOverrule


    'Test function
    Public Overrides Function DeepClone(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal ownerObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal idMap As Autodesk.AutoCAD.DatabaseServices.IdMapping, ByVal isPrimary As Boolean) As Autodesk.AutoCAD.DatabaseServices.DBObject
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("ObjectOverrule.Deeclone called" & vbLf)

        If idMap.DeepCloneContext = DeepCloneType.Copy Then
            HelperClass.GetSingleton.isCopying = True
        End If

        Dim rb As ResultBuffer = HelperClass.GetSingleton.GetXrecordData(dbObject, HelperClass.GetSingleton.XRecordName2)
        Dim objId As ObjectId = rb.AsArray(0).Value
        'objId is now the ObjectId we stored in the Line's XRecord

        If Not rb Is Nothing Then
            'if we don't find the ObjId of our linked entity in the idMap, then we must clone it it
            If Not idMap.Contains(objId) Then
                Using tr As OpenCloseTransaction = dbObject.Database.TransactionManager.StartOpenCloseTransaction
                    Dim linkedObj As DBObject = tr.GetObject(objId, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForRead)
                    Dim linkedObjClone As DBObject = linkedObj.DeepClone(ownerObject, idMap, isPrimary)
                    tr.AddNewlyCreatedDBObject(linkedObjClone, True)
                    tr.Commit()
                End Using
            End If
        End If

        'Now clone our line
        Return MyBase.DeepClone(dbObject, ownerObject, idMap, isPrimary)
    End Function



 



    Public Overrides Function WblockClone(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal ownerObject As Autodesk.AutoCAD.Runtime.RXObject, ByVal idMap As Autodesk.AutoCAD.DatabaseServices.IdMapping, ByVal isPrimary As Boolean) As Autodesk.AutoCAD.DatabaseServices.DBObject
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("ObjectOverrule.WblockClone called" & vbLf)


        Dim rb As ResultBuffer = HelperClass.GetSingleton.GetXrecordData(dbObject, HelperClass.GetSingleton.XRecordName2)
        Dim objId As ObjectId = rb.AsArray(0).Value
        'objId is now the ObjectId we stored in the Line's XRecord

        If Not rb Is Nothing Then
            'if we don't find the ObjId of our linked entity in the idMap, then we must clone it it
            If Not idMap.Contains(objId) Then
                Using tr As OpenCloseTransaction = dbObject.Database.TransactionManager.StartOpenCloseTransaction
                    Dim linkedObj As DBObject = tr.GetObject(objId, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForRead)
                    Dim linkedObjClone As DBObject = linkedObj.WblockClone(ownerObject, idMap, isPrimary)
                    tr.AddNewlyCreatedDBObject(linkedObjClone, True)
                    tr.Commit()
                End Using
            End If
        End If

        'Now clone our line
        Return MyBase.WblockClone(dbObject, ownerObject, idMap, isPrimary)

    End Function

    'Public Overrides Sub Open(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal mode As Autodesk.AutoCAD.DatabaseServices.OpenMode)
    '    Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("ObjectOverrule.Open called" & vbLf)
    '    MyBase.Open(dbObject, mode)
    'End Sub

    'Public Overrides Sub Close(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject)
    '    Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("ObjectOverrule.Close called" & vbLf)
    '    MyBase.Close(dbObject)
    'End Sub

    'Public Overrides Sub Cancel(ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject)
    '    Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("ObjectOverrule.Cancel called" & vbLf)
    '    MyBase.Cancel(dbObject)
    'End Sub

    'Public Overrides Sub [Erase](ByVal dbObject As Autodesk.AutoCAD.DatabaseServices.DBObject, ByVal erasing As Boolean)
    '    Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("ObjectOverrule.[Erase] called" & vbLf)
    '    MyBase.[Erase](dbObject, erasing)
    'End Sub


End Class

#End Region



#Region "Osnap Overrule"
Public Class MyOsnapOverrule
    Inherits OsnapOverrule

    Public Overrides Sub GetObjectSnapPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal snapMode As Autodesk.AutoCAD.DatabaseServices.ObjectSnapModes, ByVal gsSelectionMark As System.IntPtr, ByVal pickPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal lastPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal viewTransform As Autodesk.AutoCAD.Geometry.Matrix3d, ByVal snapPoints As Autodesk.AutoCAD.Geometry.Point3dCollection, ByVal geometryIds As Autodesk.AutoCAD.Geometry.IntegerCollection)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("OsnapOverrule.GetObjectSnapPoints called" & vbLf)
        MyBase.GetObjectSnapPoints(entity, snapMode, gsSelectionMark, pickPoint, lastPoint, viewTransform, snapPoints, geometryIds)
    End Sub

    Public Overrides Sub GetObjectSnapPoints(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal snapMode As Autodesk.AutoCAD.DatabaseServices.ObjectSnapModes, ByVal gsSelectionMark As System.IntPtr, ByVal pickPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal lastPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal viewTransform As Autodesk.AutoCAD.Geometry.Matrix3d, ByVal snapPoints As Autodesk.AutoCAD.Geometry.Point3dCollection, ByVal geometryIds As Autodesk.AutoCAD.Geometry.IntegerCollection, ByVal insertionMat As Autodesk.AutoCAD.Geometry.Matrix3d)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("OsnapOverrule.GetObjectSnapPoints called" & vbLf)
        MyBase.GetObjectSnapPoints(entity, snapMode, gsSelectionMark, pickPoint, lastPoint, viewTransform, snapPoints, geometryIds, insertionMat)
    End Sub

End Class
#End Region


#Region "Subentity Overrule"

Public Class MySubentityOverrule
    Inherits SubentityOverrule

    Public Overrides Sub AddSubentPaths(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal paths() As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.AddSubentPaths called" & vbLf)
        MyBase.AddSubentPaths(entity, paths)
    End Sub

    Public Overrides Sub DeleteSubentPaths(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal paths() As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.DeleteSubentPaths called" & vbLf)
        MyBase.DeleteSubentPaths(entity, paths)
    End Sub

    Public Overrides Function GetCompoundObjectTransform(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity) As Autodesk.AutoCAD.Geometry.Matrix3d
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.GetCompoundObjectTransform called" & vbLf)
        Return MyBase.GetCompoundObjectTransform(entity)
    End Function

    Public Overrides Sub GetGripPointsAtSubentPath(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal subPath As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath, ByVal grips As Autodesk.AutoCAD.DatabaseServices.GripDataCollection, ByVal curViewUnitSize As Double, ByVal gripSize As Integer, ByVal curViewDir As Autodesk.AutoCAD.Geometry.Vector3d, ByVal bitFlags As Autodesk.AutoCAD.DatabaseServices.GetGripPointsFlags)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.GetGripPointsAtSubentPath called" & vbLf)
        MyBase.GetGripPointsAtSubentPath(entity, subPath, grips, curViewUnitSize, gripSize, curViewDir, bitFlags)
    End Sub

    Public Overrides Function GetGsMarkersAtSubentPath(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal subPath As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath) As System.IntPtr()
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.GetGsMarkersAtSubentPath called" & vbLf)
        Return MyBase.GetGsMarkersAtSubentPath(entity, subPath)
    End Function

    Public Overrides Function GetSubentClassId(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal path As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath) As System.Guid
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.GetSubentClassId called" & vbLf)
        Return MyBase.GetSubentClassId(entity, path)
    End Function

    Public Overrides Function GetSubentPathGeomExtents(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal path As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath) As Autodesk.AutoCAD.DatabaseServices.Extents3d
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.GetSubentPathGeomExtents called" & vbLf)
        Return MyBase.GetSubentPathGeomExtents(entity, path)
    End Function

    Public Overrides Function GetSubentPathsAtGsMarker(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal type As Autodesk.AutoCAD.DatabaseServices.SubentityType, ByVal gsMark As System.IntPtr, ByVal pickPoint As Autodesk.AutoCAD.Geometry.Point3d, ByVal viewXform As Autodesk.AutoCAD.Geometry.Matrix3d, ByVal entAndInsertStack() As Autodesk.AutoCAD.DatabaseServices.ObjectId) As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath()
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.GetSubentPathsAtGsMarker called" & vbLf)
        Return MyBase.GetSubentPathsAtGsMarker(entity, type, gsMark, pickPoint, viewXform, entAndInsertStack)
    End Function

    Public Overrides Sub MoveGripPointsAtSubentPaths(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal path() As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath, ByVal grips As Autodesk.AutoCAD.DatabaseServices.GripDataCollection, ByVal offset As Autodesk.AutoCAD.Geometry.Vector3d, ByVal bitFlags As Autodesk.AutoCAD.DatabaseServices.MoveGripPointsFlags)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.MoveGripPointsAtSubentPaths called" & vbLf)
        MyBase.MoveGripPointsAtSubentPaths(entity, path, grips, offset, bitFlags)
    End Sub

    Public Overrides Sub OnSubentGripStatusChanged(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal status As Autodesk.AutoCAD.DatabaseServices.GripStatus, ByVal subentity As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.OnSubentGripStatusChanged called" & vbLf)
        MyBase.OnSubentGripStatusChanged(entity, status, subentity)
    End Sub

    Public Overrides Function SubentPtr(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal id As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath) As Autodesk.AutoCAD.DatabaseServices.Entity
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.SubentPtr called" & vbLf)
        Return MyBase.SubentPtr(entity, id)
    End Function

    Public Overrides Sub TransformSubentPathsBy(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal paths() As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath, ByVal xform As Autodesk.AutoCAD.Geometry.Matrix3d)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("SubentityOverrule.TransformSubentPathsBy called" & vbLf)
        MyBase.TransformSubentPathsBy(entity, paths, xform)
    End Sub

End Class



#End Region


#Region "Highlight Overrule"

Public Class MyHighlightOverrule
    Inherits HighlightOverrule

    Public Overrides Sub Highlight(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal subId As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath, ByVal highlightAll As Boolean)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("HighlightOverrule.Highlight called" & vbLf)
        MyBase.Highlight(entity, subId, highlightAll)
    End Sub

    Public Overrides Sub Unhighlight(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal subId As Autodesk.AutoCAD.DatabaseServices.FullSubentityPath, ByVal highlightAll As Boolean)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("HighlightOverrule.Unhighlight called" & vbLf)
        MyBase.Unhighlight(entity, subId, highlightAll)
    End Sub

End Class

#End Region



#Region "Geometry Overrule"

Public Class MyGeometryOverrule
    Inherits GeometryOverrule

    Public Overrides Function GetGeomExtents(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity) As Autodesk.AutoCAD.DatabaseServices.Extents3d
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GeometryOverrule.GetGeomExtents called" & vbLf)

        'Start with extents of original object
        Dim myExtents As Extents3d = MyBase.GetGeomExtents(entity)

        'If this isn't a line (and it should be) we return original entity's extents
        If Not TypeOf entity Is Line Then
            Return myExtents
        End If

        'Collect explode fragments and calculate total extents from extents of individual 
        Using entSet As New DBObjectCollection
            HelperClass.GetSingleton.ThermometerExplodeFragments(entity, entSet)
            For Each ent As Entity In entSet
                myExtents.AddExtents(ent.GeometricExtents)
            Next
        End Using

        Return myExtents

    End Function

    Public Overrides Sub IntersectWith(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal ent As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal intType As Autodesk.AutoCAD.DatabaseServices.Intersect, ByVal projPlane As Autodesk.AutoCAD.Geometry.Plane, ByVal points As Autodesk.AutoCAD.Geometry.Point3dCollection, ByVal thisGsMarker As System.IntPtr, ByVal otherGsMarker As System.IntPtr)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GeometryOverrule.IntersectWith1 called" & vbLf)

        If Not TypeOf entity Is Line Then
            MyBase.IntersectWith(entity, ent, intType, projPlane, points, thisGsMarker, otherGsMarker)
        End If

        'Collect explode fragments and use returned entity set to calc intersections whole overruled entity
        Using entSet As New DBObjectCollection
            HelperClass.GetSingleton.ThermometerExplodeFragments(entity, entSet)
            For Each subEnt As Entity In entSet
                subEnt.IntersectWith(ent, intType, projPlane, points, thisGsMarker, otherGsMarker)
            Next
        End Using

        'Not needed because we make a 'copy' of the original line in the explode fragments set
        'MyBase.IntersectWith(entity, ent, intType, projPlane, points, thisGsMarker, otherGsMarker)

    End Sub

    Public Overrides Sub IntersectWith(ByVal entity As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal ent As Autodesk.AutoCAD.DatabaseServices.Entity, ByVal intType As Autodesk.AutoCAD.DatabaseServices.Intersect, ByVal points As Autodesk.AutoCAD.Geometry.Point3dCollection, ByVal thisGsMarker As System.IntPtr, ByVal otherGsMarker As System.IntPtr)
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("GeometryOverrule.IntersectWith2 called" & vbLf)

        If Not TypeOf entity Is Line Then
            MyBase.IntersectWith(entity, ent, intType, points, thisGsMarker, otherGsMarker)
        End If

        'Explode entity and use returned entity set to calc intersections whole overruled entity
        Using entSet As New DBObjectCollection
            HelperClass.GetSingleton.thermometerexplodefragments(entity, entSet)
            For Each subEnt As Entity In entSet
                subEnt.IntersectWith(ent, intType, points, thisGsMarker, otherGsMarker)
            Next
        End Using

        'Not needed because we make a 'copy' of the original line in the explodefragments set
        'MyBase.IntersectWith(entity, ent, intType, points, thisGsMarker, otherGsMarker)

    End Sub
End Class

#End Region





#Region "Implementation of  the commands"

Public Class TestOverrule
    Implements IExtensionApplication

    'Setup some global variables
    Shared mOverrules As System.Collections.Generic.List(Of Overrule)

    'Called when DLL is loaded by AutoCAD.
    Public Sub Initialize() Implements Autodesk.AutoCAD.Runtime.IExtensionApplication.Initialize

        'Remind user what the commands are
        Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
        ed.WriteMessage(vbCrLf + "Overrule API example")
        ed.WriteMessage(vbCrLf + "Commands are:")
        ed.WriteMessage(vbCrLf + "TOGGLEOVERRULE - turns overrule protocol on and off.")
        ed.WriteMessage(vbCrLf + "ADDDATA - adds extension dictionary to selected line, and filters on Extension dictionary.")
        ed.WriteMessage(vbCrLf + "LINKENTITY - Links to another entity via XRecord data.")

        mOverrules = New System.Collections.Generic.List(Of Overrule)

        mOverrules.Add(New MyDrawOverrule)
        mOverrules.Add(New MyGripOverrule)
        mOverrules.Add(New MyTransformOverrule)
        mOverrules.Add(New MyPropertiesOverrule)
        mOverrules.Add(New MyObjectOverrule)
        mOverrules.Add(New MyOsnapOverrule)
        mOverrules.Add(New MySubentityOverrule)
        mOverrules.Add(New MyHighlightOverrule)
        mOverrules.Add(New MyGeometryOverrule)

        'Instantiate our global Overrules and set them to overrule lines with my data attached
        For Each thisOverrule As Overrule In mOverrules

            Overrule.AddOverrule(RXObject.GetClass(GetType(Line)), thisOverrule, False)
            thisOverrule.SetExtensionDictionaryEntryFilter(HelperClass.GetSingleton.DictionaryName)

        Next

        'Turn overruling on
        Overrule.Overruling = True

    End Sub

    'Clean up after ourselves.
    Public Sub Terminate() Implements Autodesk.AutoCAD.Runtime.IExtensionApplication.Terminate
        'Remove all overrules
        'Not strictly necessary because we can't unload the .NET DLL, so this is only called as AutoCAD is shutting down anyway.
        For i As Integer = 1 To mOverrules.Count
            Overrule.RemoveOverrule(RXObject.GetClass(GetType(Line)), mOverrules.Item(mOverrules.Count - 1))
            mOverrules.Remove(mOverrules.Item(mOverrules.Count - 1))
        Next
    End Sub


    'Toggles all overrules on and off.
    <CommandMethod("TOGGLEOVERRULE")> _
    Public Sub ToggleOverrule()
        Overrule.Overruling = Not Overrule.Overruling
        Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage(vbCrLf & "*** Overrule is now " & Overrule.Overruling.ToString & " ***" & vbCrLf)
        Application.DocumentManager.MdiActiveDocument.Editor.Regen()
    End Sub



    'Demo of Extension Dictionary filter. 
    'There's also an Xdata filter, but we won't demonstrate it here - its basically the same).
    'This command needs tidying up to use HelperClass functions for XData access. (Currently does its own thing).
    <CommandMethod("ADDDATA")> _
    Public Sub AddXDictFilter()

        'Select a line
        Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
        Dim opts As New PromptEntityOptions(vbCrLf + "Select a line to add Extension dictionary to:")
        opts.SetRejectMessage(vbCrLf + "Sorry dude! That's not a line" + vbCrLf)
        opts.AddAllowedClass(GetType(Line), True)
        Dim res As PromptEntityResult = ed.GetEntity(opts)

        'Only continue if a line was selected 
        If res.Status <> PromptStatus.OK Then Exit Sub

        'Open circle and make sure it has our dictionary in its extension dictionary
        Dim objId As ObjectId = res.ObjectId
        Dim db As Database = objId.Database

        Using tr As Transaction = db.TransactionManager.StartTransaction

            Dim ent As Entity = tr.GetObject(objId, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForWrite)

            'Does it have XRecord already attached? If so use this to set default values. Otherwise we need to create one later.
            Dim rb As ResultBuffer = HelperClass.GetSingleton.GetXRecordData(ent, HelperClass.GetSingleton.XRecordName)
            'rb will be null if there was no existing XRecord


            'Set some default values ready for prompting for new data
            Dim temp1Opts As New PromptIntegerOptions(vbCrLf + "Enter Lower Temperature:")
            Dim temp2Opts As New PromptIntegerOptions(vbCrLf + "Enter Upper Temperature:")
            Dim temp3Opts As New PromptIntegerOptions(vbCrLf + "Enter Current Temperature:")
            Dim nameOpts As New PromptStringOptions(vbCrLf + "Enter Name:")
            temp1Opts.LowerLimit = 0
            temp1Opts.UpperLimit = 100
            temp2Opts.LowerLimit = 0
            temp2Opts.UpperLimit = 100
            temp3Opts.LowerLimit = 0
            temp1Opts.UpperLimit = 100
            nameOpts.AllowSpaces = True


            'If our data isn't in in extension dictionary we add it. 
            If rb Is Nothing Then
                temp1Opts.DefaultValue = 20
                temp2Opts.DefaultValue = 30
                temp3Opts.DefaultValue = 25
                nameOpts.DefaultValue = "San Rafael"

                rb = New ResultBuffer( _
                        New TypedValue(DxfCode.Text, nameOpts.DefaultValue), _
                        New TypedValue(DxfCode.Int32, temp1Opts.DefaultValue), _
                        New TypedValue(DxfCode.Int32, temp2Opts.DefaultValue), _
                        New TypedValue(DxfCode.Int32, temp3Opts.DefaultValue))

                HelperClass.GetSingleton.SetXRecordData(ent, rb, HelperClass.GetSingleton.XRecordName)

            End If


            'Read our data values to populate prompt defaults
            Dim arr() As TypedValue = rb.AsArray
            Dim val1 As TypedValue = arr(0) 'Room name
            Dim val2 As TypedValue = arr(1) 'Min temp
            Dim val3 As TypedValue = arr(2) 'Max temp
            Dim val4 As TypedValue = arr(3) 'Current temp


            nameOpts.DefaultValue = val1.Value
            temp1Opts.DefaultValue = val2.Value
            temp2Opts.DefaultValue = val3.Value
            temp3Opts.DefaultValue = val4.Value


            'Prompt for new values
            Dim nameRes As PromptResult = ed.GetString(nameOpts)
            If nameRes.Status = PromptStatus.OK Then
                val1 = New TypedValue(DxfCode.Text, nameRes.StringResult)
            End If

            Dim temp1Res As PromptIntegerResult = ed.GetInteger(temp1Opts)
            If temp1Res.Status = PromptStatus.OK Then
                val2 = New TypedValue(DxfCode.Int32, temp1Res.Value)
            End If

            Dim temp2Res As PromptIntegerResult = ed.GetInteger(temp2Opts)
            If temp2Res.Status = PromptStatus.OK Then
                val3 = New TypedValue(DxfCode.Int32, temp2Res.Value)
            End If

            Dim temp3Res As PromptIntegerResult = ed.GetInteger(temp3Opts)
            If temp3Res.Status = PromptStatus.OK Then
                val4 = New TypedValue(DxfCode.Int32, temp3Res.Value)
            End If

            'Now set Xrecord contents to new values
            rb = New ResultBuffer(val1, val2, val3, val4)
            HelperClass.GetSingleton.SetXRecordData(ent, rb, HelperClass.GetSingleton.XRecordName)


            tr.Commit()



        End Using


        'Display new results
        ed.Regen()

    End Sub


    <CommandMethod("LINKENTITY")> _
    Public Sub LinkEntity()

        'Select a line
        Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor
        Dim opts As New PromptEntityOptions(vbCrLf + "Select a line to link entity to:")
        opts.SetRejectMessage(vbCrLf + "Sorry dude! That's not a line" + vbCrLf)
        opts.AddAllowedClass(GetType(Line), True)
        Dim res As PromptEntityResult = ed.GetEntity(opts)

        'Only continue if a line was selected 
        If res.Status <> PromptStatus.OK Then Exit Sub

        Dim objId As ObjectId = res.ObjectId
        Dim db As Database = objId.Database


        'Prompt for an entity to link to this thermometer.
        Dim opts2 As New PromptEntityOptions(vbCrLf + "Select entity to link:")
        res = ed.GetEntity(opts2)

        'Only continue if a line was selected 
        If res.Status <> PromptStatus.OK Then Exit Sub

        Dim val As New TypedValue(DxfCode.HardPointerId, res.ObjectId)
        Dim rb As New ResultBuffer(val)

        'Add linked entity data to line (adding ObjectId of entity we selected)
        Using tr As Transaction = db.TransactionManager.StartTransaction

            Dim ent As Entity = tr.GetObject(objId, Autodesk.AutoCAD.DatabaseServices.OpenMode.ForWrite)

            'Just set new data. We only want to link to one entity, so we don't care if we overwrite old data
            HelperClass.GetSingleton.SetXRecordData(ent, New ResultBuffer(val), HelperClass.GetSingleton.XRecordName2)

            tr.Commit()
        End Using

    End Sub



End Class


#End Region
